"""Conftest.py is loaded for each pytest. Contains fixtures shared by multiple tests
"""
import shutil
from pathlib import Path
from typing import List
from unittest.mock import Mock

import pytest
from anonapi.responses import JobStatus
from anonapi.testresources import (
    RemoteAnonServerFactory,
    JobInfoFactory,
    MockAnonClientTool,
)

from idissend.core import Stream, Stage, Study
from idissend.persistence import IDISSendRecords, get_memory_only_sessionmaker
from idissend.stages import CoolDown, PendingAnon, IDISConnection
from tests import RESOURCE_PATH
from tests.factories import StreamFactory


@pytest.fixture
def some_studies(an_incoming_stage) -> List[Study]:
    """some studies in the incoming stage with some actual data on disk"""
    return an_incoming_stage.get_all_studies()


@pytest.fixture
def a_study(some_studies) -> Study:
    """A study in the incoming stage with some actual data on disk"""
    return some_studies[0]


@pytest.fixture
def an_idssend_structured_folder(tmpdir) -> Path:
    """A folder with some DICOM files in <stream>/<study> structure"""

    copy_of_folder = Path(str(tmpdir)) / "an_idssend_structured_folder"
    shutil.copytree(RESOURCE_PATH / "an_idssend_structured_folder", copy_of_folder)
    return copy_of_folder


@pytest.fixture
def some_streams() -> List[Stream]:
    """Some streams, some of which have some data in an_idssend_structured_folder()"""
    return [
        StreamFactory(name="project1"),
        StreamFactory(name="project2"),
        StreamFactory(name="project3"),
    ]


@pytest.fixture
def some_stages(some_streams, tmpdir) -> List[Stage]:
    """Two stages that each have some streams and an empty tmp path"""

    return [
        Stage(name="stage1", streams=some_streams, path=Path(tmpdir) / "stage1"),
        Stage(name="stage2", streams=some_streams, path=Path(tmpdir) / "stage2"),
    ]


@pytest.fixture
def a_stage(some_stages):
    """A stage with some streams and an empty tmp path"""
    return some_stages[0]


@pytest.fixture
def an_incoming_stage(an_idssend_structured_folder, some_streams) -> CoolDown:
    """An incoming stage which has some actual content on disk"""
    return CoolDown(
        name="incoming",
        path=an_idssend_structured_folder,
        streams=some_streams,
        cool_down=0,
    )


@pytest.fixture()
def an_empty_pending_stage(
    some_streams, an_idis_connection, tmpdir, a_records_db
) -> PendingAnon:
    """An empty pending stage with a mocked connection to IDIS and mocked records db
    """
    return PendingAnon(
        name="pending",
        path=Path(tmpdir) / "pending_anon",
        streams=some_streams,
        idis_connection=an_idis_connection,
        records=a_records_db,
    )


@pytest.fixture
def an_idis_connection(mock_anon_client_tool):
    """An idis connection that mocks repsonses and does not hit any server"""
    return IDISConnection(
        client_tool=mock_anon_client_tool,
        servers=[RemoteAnonServerFactory(), RemoteAnonServerFactory()],
    )


@pytest.fixture
def a_records_db() -> IDISSendRecords:
    """An initialised empty records database"""
    return IDISSendRecords(get_memory_only_sessionmaker())


@pytest.fixture
def mock_anon_client_tool(monkeypatch):
    """An anonymization API client tool that does not hit the server but returns
    some example responses instead. Also records calls"""
    some_responses = [
        JobInfoFactory(status=JobStatus.DONE),
        JobInfoFactory(status=JobStatus.ERROR),
        JobInfoFactory(status=JobStatus.INACTIVE),
    ]
    # mock wrapper to be able to record responses
    return Mock(wraps=MockAnonClientTool(responses=some_responses))



""" core and stages define the pieces of a pipeline,
this module has classes and methods for the relationships between them
"""


import logging

from anonapi.client import AnonClientTool
from anonapi.objects import RemoteAnonServer
from anonapi.paths import UNCMapping, UNCMap, UNCPath
from anonapi.responses import JobStatus
from collections import Counter
from idissend.core import Stream, Person, Stage
from idissend.persistence import IDISSendRecords, get_db_sessionmaker
from idissend.stages import CoolDown, PendingAnon, IDISConnection, Trash
from pathlib import Path

class DefaultPipeline:
    """Standard version of a idissend pipeline: data comes in, is exposed to IDIS for
    anonymization, then ends up in completed. There are trash and errorred stages to
    hold studies if needed

    Responsibilities
    ----------------
    Pipeline does:
    * Inspect status and studies of stages
    * Push studies between stages if needed
    * log errors raised by stages, but not necessarily catch

    Pipeline does NOT:
    * Handle any actual files (this is up to the stage)
    * Handle any rollback on error (this is also up to the stage)

    """

    def __init__(
        self,
        incoming: CoolDown,
        pending: PendingAnon,
        finished: CoolDown,
        trash: Trash,
        errored: Stage,
    ):
        """

        Parameters
        ----------
        incoming: CoolDown
            Data comes in here
        pending: PendingAnon
            Data waits here to be downloaded and anonymized by IDIS
        finished: CoolDown
            When IDIS is done, move the data from pending here
        trash: Trash
            Any study here can be removed when needed
        errored:
            Holds studies with errors, either in IDIS or in pipeline itself
        """
        self.incoming = incoming
        self.pending = pending
        self.finished = finished
        self.trash = trash
        self.errored = errored
        self.all_stages = [incoming, pending, finished, trash, errored]
        self.logger = logging.getLogger(__name__)

    def run_once(self):
        """Check each stage and perform actions one time

        Raises
        ------
        IDISSendException:
            If anything goes wrong executing the run that the pipeline cannot handle
            by itself
        """

        self.logger.info("Running once")

        studies = self.pending.get_all_studies()
        self.logger.debug(f"Updating IDIS status for {len(studies)} pending jobs")
        self.pending.update_records(studies)
        self.logger.debug(
            f"Found {str(dict(Counter([x.last_status for x in studies])))}."
            f" Taking action based on status"
        )

        self.finished.push_studies(
            [x for x in studies if x.last_status == JobStatus.DONE]
        )
        self.trash.push_studies(
            [x for x in studies if x.last_status == JobStatus.INACTIVE]
        )
        self.errored.push_studies(
            [x for x in studies if x.last_status == JobStatus.ERROR]
        )

        self.logger.debug("Checking for new studies coming in")
        cooled_down = self.incoming.get_all_cooled_studies()
        self.logger.debug(f"Found {len(cooled_down)}. Pushing to pending")
        self.pending.push_studies(cooled_down)

        self.logger.debug("empty old trash")
        self.trash.delete_all()

        self.logger.debug(
            f"Moving finished studies older than "
            f"{self.finished.cool_down} minutes to trash"
        )
        self.trash.push_studies(self.finished.get_all_cooled_studies())

    def get_status(self) -> str:
        """Status for all stages"""
        status_lines = []
        for stage in self.all_stages:
            studies = stage.get_all_studies()
            status_lines.append(
                f"{stage.name} contains {len(studies)} "
                f"studies: {[str(x) for x in studies]}"
            )

        return "\n".join(status_lines)



"""Specific implementations of stages the data goes through"""
import shutil

from anonapi.client import AnonClientTool, ClientToolException
from anonapi.exceptions import AnonAPIException
from anonapi.objects import RemoteAnonServer
from anonapi.paths import UNCMapping, UNCMappingException
from anonapi.responses import JobsInfoList
from collections import defaultdict
from datetime import datetime
from idissend.core import Stage, Stream, Study, PushStudyCallbackException
from idissend.exceptions import IDISSendException
from idissend.orm import PendingAnonRecord
from idissend.persistence import IDISSendRecords
from pathlib import Path
from sqlalchemy.exc import SQLAlchemyError
from typing import List, Optional, Iterable


class CoolDown(Stage):
    """A stage with an inbuilt waiting or cool down period.

    By default studies are only returned after they have been in the stage for this
    period. There are two obvious use cases for this:

    * Making sure a DICOM transfer is complete before starting processing. DICOM
      CSTORE infamously does not have a concept of 'all files have been sent'.
      In the dicom world any file could come in at any time. A common hack to
      get around this is to assume that if no files have been added for x minutes
      the transfer is probably complete.
    * Removing completed and errored studies after a certain grace period.

    """

    def __init__(
        self, name: str, path: Path, streams: List[Stream], cool_down: int = 5
    ):
        """

        Parameters
        ----------
        name: str
            Human readable name for this stage
        path: str
            Root path of this folder
        streams: List[Stream]
            All the streams for which this folder could receive data
        cool_down: int, optional
            Number of minutes to wait before considering a study cooled down.
            CoolDown is measured against the last modification date of any file
            in the study. Defaults to 5 minutes
        """
        super(CoolDown, self).__init__(name=name, path=path, streams=streams)
        self.cool_down = cool_down

    def get_all_studies(self) -> List[Study]:
        """Get all studies for all streams in this folder

        """
        return super(CoolDown, self).get_all_studies()

    def get_all_cooled_studies(self) -> List[Study]:
        """Get all studies which have not changed in the cool down period

        """
        return [x for x in self.get_all_studies() if self.has_cooled_down(x)]

    def has_cooled_down(self, study: Study) -> bool:
        """Check whether files are still coming in for this study

        Considered cooled down if no file was modified less then <cool_down> mins ago
        """
        return study.is_older_than(self.cool_down)


class IDISConnection:
    """Everything you need to talk to the IDIS anonymization server"""

    def __init__(self, client_tool: AnonClientTool, servers: List[RemoteAnonServer]):
        self.client_tool = client_tool
        self.servers = servers

    def __str__(self):
        return f"Connection with IDIS servers {[str(x) for x in self.servers]}"

    def get_server(self, server_name) -> RemoteAnonServer:
        """Find the IDIS server with the given name

        Raises
        ------
        UnknownServerException
            When no server can be found with that name
        """
        server = {x.name: x for x in self.servers}.get(server_name)
        if server:
            return server
        else:
            raise UnknownServerException(
                f"Server '{server_name}' not found. Known s"
                f"ervers:{[str(x) for x in self.servers]}"
            )


class PendingStudy(Study):
    """A study pending anonymization. Has additional links to check its status"""

    def __init__(
        self, name: str, stream: Stream, stage: Stage, record: PendingAnonRecord
    ):
        super(PendingStudy, self).__init__(name, stream, stage)
        self.record = record

    @property
    def last_status(self) -> str:
        """Get last known status of IDIS job for this study from records

        Returns
        -------
        str
            One of anonapi.responses.JobStatus
        """
        return self.record.last_status

    @property
    def last_check(self) -> Optional[datetime]:
        """Date when the job for this study was last updated"""
        return self.record.last_check

    @property
    def job_id(self) -> int:
        """IDIS job id for the job associated with this study"""
        return self.record.job_id

    @property
    def server_name(self) -> str:
        return self.record.server_name


class PendingAnon(Stage):
    """Stage where data is presented to IDIS for anonymization. Monitors progress
    of anonymization and removes data when anonymization is done or failed.

    * Caches job status in a local db
    * Communicates with IDIS to get info if needed
    * Returns PendingStudy objects instead of Study
    """

    def __init__(
        self,
        name: str,
        path: Path,
        streams: List[Stream],
        idis_connection: IDISConnection,
        records: IDISSendRecords,
        unc_mapping: UNCMapping = None,
    ):
        """

        Parameters
        ----------
        name: str
            Human readable name for this stage
        path: str
            Root path of this folder
        streams: List[Stream]
            All the streams for which this folder could receive data
        idis_connection: IDISConnection
            Used for communicating with IDIS
        records: IDISSendRecords
            Persistent storage for which studies have been sent to IDIS
        unc_mapping: UNCMapping, optional
            Translates any local paths to their UNC equivalents, making sure
            only UNC paths get sent to IDIS. If not given, use any paths as-is
        """

        super(PendingAnon, self).__init__(name=name, path=path, streams=streams)
        self.idis_connection = idis_connection
        self.records = records
        self.unc_mapping = unc_mapping

    def push_study_callback(self, study: Study) -> PendingStudy:
        """Creates an IDIS job for the given study

        Parameters
        ----------
        study: Study
            Study that was just pushed

        Raises
        ------
        PushStudyCallbackException
            When anything goes wrong executing this callback

        Returns
        -------
        PendingStudy
            The study after executing the callback

        """

        # Create a job
        client = self.idis_connection.client_tool
        server = self.idis_connection.servers[0]
        source_path = study.get_path()
        destination_path = study.stream.output_folder
        if self.unc_mapping:
            try:
                source_path = self.unc_mapping.to_unc(source_path)
                destination_path = self.unc_mapping.to_unc(destination_path)
            except UNCMappingException as e:
                raise PushStudyCallbackException(e)

        try:
            created = client.create_path_job(
                server=server,
                project_name=str(study.stream.idis_profile),
                source_path=source_path,
                destination_path=destination_path,
                description=f"Created by idissend for stream " f"{study.stream}",
                pims_keyfile_id=study.stream.pims_key,
            )
            self.logger.info(
                f"Created IDIS job {created.job_id} on {server} " f"for {study}"
            )
        except AnonAPIException as e:
            raise PushStudyCallbackException(e)

        # save job id for this study to check back on later
        try:
            with self.records.get_session() as session:
                record = session.add(
                    study_folder=study.get_path(),
                    job_id=created.job_id,
                    server_name=server.name,
                )

        except SQLAlchemyError as e:
            raise PushStudyCallbackException(e)

        return self.to_pending_study(study=study, record=record)

    def get_all_studies(self) -> List[PendingStudy]:
        """PendingAnon returns PendingStudy objects, which hold additional info on
        IDIS job status """
        studies = super().get_all_studies()
        return [self.to_pending_study(x) for x in studies]

    def get_studies(self, stream: Stream) -> List[PendingStudy]:
        """Get all studies for the given stream

        """
        return [self.to_pending_study(x) for x in super().get_studies(stream=stream)]

    def to_pending_study(
        self, study: Study, record: PendingAnonRecord = None
    ) -> PendingStudy:
        """Combine Study with a record, turning into PendingStudy. If record
        is not given, look it up in records db

        Parameters
        ----------
        study: Study
            Study object to turn onto PendingStudy
        record: PendingStudy, optional
            The record to include with this study. If not given, look for
            record based on study path

        Raises
        ------
        RecordNotFoundException
            If the given study has no record in the records database
        """
        if not record:
            with self.records.get_session() as session:
                record = session.get_for_study_folder(study.get_path())
        if not record:
            raise RecordNotFoundException(
                f"{str(self)}: There is no record for {study}"
            )
        return PendingStudy(
            name=study.name, stream=study.stream, stage=study.stage, record=record
        )

    def get_all_orphaned_studies(self) -> List[Study]:
        """Returns all studies for which no records exists.

        This should not occur often but could be the result of certain crashes or if
        data has been inserted into streams from outside idissend. This is the
        only method in this stage which does not return PendingStudy objects.

        Returns
        -------
        List[Study]
            All studies for which no record exists.
        """

        with self.records.get_session() as session:
            records = session.get_all()

        studies = super(PendingAnon, self).get_all_studies()
        study_paths = {x.get_path(): x for x in studies}
        record_paths = set((x.study_folder for x in records))
        return [y for x, y in study_paths.items() if x not in record_paths]

    def get_server(self, server_name: str) -> RemoteAnonServer:
        return self.idis_connection.get_server(server_name=server_name)

    def update_records(self, studies: List[PendingStudy]) -> List[PendingStudy]:
        """Contact IDIS to get updated status for all given studies

        Raises
        ------
        IDISCommunicationException
            If anything goes wrong getting information from IDIS
        """
        # group jobs per IDIS server to minimize number of web API queries
        studies_per_server = defaultdict(list)
        for study in studies:
            studies_per_server[self.get_server(study.server_name)].append(study)

        job_info_per_study = {}
        for server in studies_per_server:
            # get info from IDIS
            job_infos = self.get_job_info_list(
                server=server, job_ids=[x.job_id for x in studies_per_server[server]]
            )

            # associate what you get back with studies
            for study in studies_per_server[server]:
                # job infos might nog be in the same order, find them again
                info = {x.job_id: x for x in job_infos}.get(study.job_id)
                if not info:
                    raise IDISCommunicationException(
                        f"study '{study}' should have a job with id {study.job_id} "
                        f"in {server}, but that job id does not seem to exist there"
                    )
                job_info_per_study[study] = info

        # now update all records with the gathered IDIS data
        with self.records.get_session() as session:
            for study, job_info in job_info_per_study.items():
                study.record.last_status = job_info.status
                study.record.last_check = datetime.now()
                session.add_record(study.record)

        return list(job_info_per_study.keys())

    def get_job_info_list(
        self, server: RemoteAnonServer, job_ids: Iterable[int]
    ) -> JobsInfoList:
        """Contact IDIS server for updated info given jobs

        Raises
        ------
        IDISCommunicationException
            If anything goes wrong communicating with IDIS

        """
        try:
            return self.idis_connection.client_tool.get_job_info_list(
                server=server, job_ids=list(job_ids)
            )
        except ClientToolException as e:
            raise IDISCommunicationException(e)


class Trash(Stage):
    """Where studies are sent after they have been anonymized.

    Can be emptied kind of prudently (Keep as much as possible but also keep
    enough space left)"""

    def delete_all(self):
        """delete data for all studies in trash"""
        studies = self.get_all_studies()
        self.logger.info(
            f"Removing data for {len(studies)} studies: " f"{[str(x) for x in studies]}"
        )
        for study in studies:
            shutil.rmtree(study.get_path())


class RecordNotFoundException(IDISSendException):
    pass


class IDISCommunicationException(IDISSendException):
    """Something went wrong getting info from IDIS"""

    pass


class UnknownServerException(IDISCommunicationException):
    pass



"""Core concepts in idissend"""
import logging
import shutil

from datetime import datetime
from idissend.exceptions import IDISSendException
from pathlib import Path
from typing import List


class Person:
    """A person with contact details"""

    def __init__(self, name: str, email: str):
        self.name = name
        self.email = email

    def __str__(self):
        return self.name


class Stream:
    """A single route that incoming data goes through.
    Determines anonymization type and destination.

    Notes
    -----
    Responsibilities: A stream is a passive data structure. It should not know about
    where the data it contains is exactly. This is the responsibility of each Stage
    that a stream is in

    """

    def __init__(
        self,
        name: str,
        output_folder: Path,
        idis_profile: str,
        pims_key: str,
        contact: Person,
    ):
        """

        Parameters
        ----------
        name: str
            Name of this stream, doubles as folder name
        output_folder: Path
            Final destination of the data. Full UNC path
        idis_profile:
            Use the settings in this project for anonymization
        pims_key:
            Use this PIMS project for generating pseudonyms
        contact:
            Who is responsible for collecting this data in the end?
        """

        self.name = name
        self.output_folder = output_folder
        self.idis_profile = idis_profile
        self.pims_key = pims_key
        self.contact = contact

    def __str__(self):
        return self.name


class IncomingFile:
    """A file which just came in to a folder. Allows checking of age()
    """

    def __init__(self, path: Path):
        self.path = path

    def __str__(self):
        return str(self.path)

    def age(self) -> float:
        """Minutes since last modification of this file

        Raises
        ------

        """
        delta = datetime.now() - datetime.fromtimestamp(self.path.stat().st_mtime)
        return delta.total_seconds() / 60


class Study:
    """A folder containing files that all belong to the same study """

    def __init__(self, name: str, stream: Stream, stage: "Stage"):
        self.name = name
        self.stream = stream
        self.stage = stage

    def __str__(self):
        return f"{self.stream}:{self.name}"

    def get_path(self) -> Path:
        """Full path to the folder that data for this study is in"""
        return self.stage.get_path_for_study(self)

    def get_files(self) -> List[Path]:
        """All files directly in this folder (no recursing)"""
        return [x for x in self.get_path().glob("*") if x.is_file()]

    def age(self) -> float:
        """Minutes since last modification of any file in this study"""
        return min(IncomingFile(x).age() for x in self.get_files())

    def is_older_than(self, minutes: float):
        """Is age of each file in this study greater than minutes?

        Potentially faster than age() as it does not check all files
        """
        for file in self.get_files():
            if IncomingFile(file).age() <= minutes:
                return False

        return True


class Stage:
    """A distinct step in the pipeline. Contains Studies for different Streams

    A stage is a something like incoming, pending, trash.
    In addition to containing data, a stage might have additional responsibilities
    like communicating with IDIS to determine the status of studies or checking
    file attributes.

    Notes
    -----
    Responsibilities:
    * A stage must know exactly where data is based on stream and study
    * A stage can do internal bookkeeping with the studies it holds
    * A stage must never push or pull studies by itself; moving data away from
      or into a stage is done from the outside.

    """

    def __init__(self, name: str, path: Path, streams: List[Stream]):
        """

        Parameters
        ----------
        name: str
            Human readable name for this stage
        path: str
            Root path of this folder
        streams: List[Stream]
            All the streams for which this folder could receive data

        """
        self.name = name
        self.path = path
        self.streams = streams
        self.logger = logging.getLogger(f'stage "{self.name}"')

    def __str__(self):
        return self.name

    def push_studies(self, studies: List[Study]) -> List[Study]:
        """Insert each study into this stage

        Raises
        ------
        StudyPushException:
            When pushing any study does not work for some reason
        """

        pushed = []
        for study in studies:
            pushed.append(self.push_study(study))

        return pushed

    def push_study(self, study: Study, stream: Stream = None) -> Study:
        """Push the given study to this stage. Optionally set stream.

        Parameters
        ----------
        study: Study
            Send the data in this study
        stream: Stream, optional
            If given, push to this stream. Otherwise use study.stream

        Raises
        ------
        StudyPushException:
            When pushing the study does not work for some reason

        Returns
        -------
        Study:
            The study after pushing to this stage. New object

        """
        self.logger.debug(f"receiving {study}")
        if not stream:
            stream = study.stream

        if stream in self.streams:
            self.assert_path_for_stream(stream)
        else:
            raise StudyPushException(f"Stream '{stream}' " f"does not exist in {self}")

        # create new study that is in this stage
        original_study = study  # keep original for possible rollback
        new_study = Study(name=study.name, stream=stream, stage=self)

        # now move the data from original to new
        try:
            shutil.move(
                str(original_study.get_path()),
                str(new_study.stage.get_path_for_stream(stream)),
            )
            return self.push_study_callback(new_study)

        except (IDISSendException, PushStudyCallbackException) as e:
            self.logger.warning(f"receiving {study} failed: {e}. Rolling back.")
            # roll back. move data back where it came from
            shutil.move(
                str(new_study.get_path()),
                str(original_study.stage.get_path_for_stream(study.stream)),
            )
            raise StudyPushException(e)
        except (FileNotFoundError, OSError) as e:
            self.logger.warning(f"receiving {study} failed: {e}")
            raise StudyPushException(e)

    def push_study_callback(self, study: Study):
        """Function that gets called directly after a study gets pushed to this stage

        Separate callback to house functionality specific to a stage. Separate
        from general functionality in push_study()

        Parameters
        ----------
        study: Study

        Raises
        ------
        PushStudyCallbackException
            When anything goes wrong executing this callback

        Returns
        -------
        Study
            The study after executing the callback

        """
        # Implement specifics in child classes
        return study

    def get_path_for_stream(self, stream: Stream) -> Path:
        """Get the folder where data is for this stream """

        return self.path / stream.name

    def assert_path_for_stream(self, stream: Stream) -> Path:
        """Create path for stream if it does not exist"""
        path = self.get_path_for_stream(stream)
        path.mkdir(parents=True, exist_ok=True)
        return path

    def get_path_for_study(self, study: Study) -> Path:
        """Get the folder where data is for this study"""

        return self.get_path_for_stream(study.stream) / study.name

    def get_all_studies(self) -> List[Study]:
        """Get all studies for all streams in this stage
        """

        studies = []
        for stream in self.streams:
            studies += self._get_studies(stream)

        return studies

    def get_studies(self, stream: Stream) -> List[Study]:
        """Get all studies for the given stream

        """
        return self._get_studies(stream=stream)

    def _get_studies(self, stream: Stream) -> List[Study]:
        """hidden method because get_studies itself can be overwritten in parent
        classes

        """
        studies = []
        for folder in [x for x in self.get_path_for_stream(stream).glob("*")]:
            studies.append(Study(name=folder.name, stream=stream, stage=self))

        return studies

    def assert_all_paths(self):
        """Make sure paths to this stage and all stream in it exist

        Useful for initial testing of a stage: you don't have to remember
        the exact paths for expected data"""
        for stream in self.streams:
            self.get_path_for_stream(stream).mkdir(parents=True, exist_ok=True)


class UnknownStreamException(IDISSendException):
    pass


class StudyPushException(IDISSendException):
    pass


class PushStudyCallbackException(IDISSendException):
    pass
